import base64
import io
import matplotlib.backend_bases
from matplotlib.backends.backend_agg import FigureCanvasAgg


class Q8SLoggerBackend(FigureCanvasAgg):
    """Custom Matplotlib backend that logs images as base64-encoded strings."""

    def print_png(self, filename_or_obj, *args, **kwargs):
        """Override the PNG printing to log Base64 instead of saving to a file."""
        buf = io.BytesIO()
        super().print_png(buf, *args, **kwargs)

        # Convert to base64
        buf.seek(0)
        encoded_image = base64.b64encode(buf.getvalue()).decode("utf-8")

        # Log to console
        print(f"data:image/png;base64,{encoded_image}\n")


FigureCanvas = Q8SLoggerBackend
# Register the backend


# Fix: Override `show()` to manually trigger rendering
def show():
    """Manually process all figures and log their base64 images."""
    import matplotlib._pylab_helpers

    # Get all active figures
    figures = matplotlib._pylab_helpers.Gcf.get_all_fig_managers()

    for manager in figures:
        figure = manager.canvas.figure
        canvas = Q8SLoggerBackend(figure)
        canvas.draw()  # Render the figure
        canvas.print_png(None)  # Trigger our Base64 logging


# Register the custom show function
matplotlib.backend_bases.show = show



import base64
from json import JSONEncoder, loads
import os
from string import Template
import tempfile
import shutil
from time import sleep
import logging
import string
import random

from dotenv import dotenv_values
from python_on_whales import docker
from kubernetes import client

from .deps.parser import Parser

FORMAT = "[%(levelname)s %(asctime)-15s q8s_kernel] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)

TAG_TEMPLATE = Template("$image:$version")
NAMESPACE = os.environ.get("NAMESPACE", "default")
MEMORY = os.environ.get("MEMORY", "32Gi")


dockerfile_content = """
FROM --platform=amd64 vstirbu/benchmark-deps

COPY main.py .

CMD ["./entrypoint.sh"]
"""

# requirements_content = """
# qiskit==1.0.0
# qiskit-aer-gpu==0.13.3
# qiskit_algorithms
# qiskit_ibm_runtime
# numpy
# networkx
# """

# entrypoint_content = """
# #!/bin/bash

# python3 ./main.py
# """


def list_containers(client):
    print(client.containers.list())


def create_temp_directory() -> str:
    temp_dir = tempfile.mkdtemp()
    logging.debug("Temporary directory created:", temp_dir)

    return temp_dir


def delete_directory(directory_path):
    shutil.rmtree(directory_path)
    print("Directory deleted:", directory_path)


def write_to_file(file_path: str, content: str):
    with open(file_path, "w") as file:
        file.write(content)
    print("Content written to file:", file_path)


def enable_executable(file_path: str):
    os.chmod(file_path, 0o755)
    print("Executable flag enabled for file:", file_path)


def prepare_build_folder(temp_dir: str, python_file_content: str):
    parser = Parser()
    deps = parser.parse(python_file_content)

    requirements_content = parser.toRequirements(deps)
    write_to_file(temp_dir + "/Dockerfile", dockerfile_content)
    write_to_file(temp_dir + "/requirements.txt", requirements_content)
    # write_to_file(temp_dir + "/entrypoint.sh", entrypoint_content)
    # write_to_file(temp_dir + "/main.py", python_file_content)
    # enable_executable(temp_dir + "/entrypoint.sh")


def whoami() -> str:
    whoami = client.AuthenticationV1Api().create_self_subject_review(body={})

    return whoami.status.user_info.username.split(":")[-1]


def load_env():
    env = dotenv_values(".env.q8s")

    # for key in env.keys():
    #     logging.info(f"{key} = {env[key]}")

    return env


def create_environment_secret(
    # job: client.V1Job,
    name: str = "app-environment",
):
    env = load_env()

    data = {}

    for key in env.keys():
        data[key] = base64.b64encode(env[key].encode()).decode()

    secret = client.V1Secret(
        api_version="v1",
        kind="Secret",
        type="Opaque",
        immutable=True,
        data=data,
        metadata=client.V1ObjectMeta(
            name=name,
            namespace=NAMESPACE,
            # owner_references=[
            #     client.V1OwnerReference(
            #         api_version="v1",
            #         kind="Job",
            #         name=job.metadata.name,
            #         uid=job.metadata.uid,
            #         # block_owner_deletion=True,
            #         # controller=True,
            #     )
            # ],
        ),
    )

    client.CoreV1Api().create_namespaced_secret(namespace=NAMESPACE, body=secret)

    logging.info("Environment created.")


def create_config_map_object(code: str, job: client.V1Job, name="app-config"):
    # Configureate ConfigMap from a local file
    configmap = client.V1ConfigMap(
        api_version="v1",
        kind="ConfigMap",
        data={"main.py": code},
        metadata=client.V1ObjectMeta(
            name=name,
            owner_references=[
                client.V1OwnerReference(
                    api_version="v1",
                    kind="Job",
                    name=job.metadata.name,
                    uid=job.metadata.uid,
                    # block_owner_deletion=True,
                    # controller=True,
                )
            ],
        ),
    )

    result = client.CoreV1Api().create_namespaced_config_map(
        namespace=NAMESPACE, body=configmap
    )

    logging.info("ConfigMap created.")


def prepare_environment(name: str):
    raw = load_env()

    env = []

    for key in raw.keys():
        env.append(
            client.V1EnvVar(
                name=key,
                value_from=client.V1EnvVarSource(
                    secret_key_ref=client.V1SecretKeySelector(name=name, key=key)
                ),
            )
        )

    return env


def registry_credentials_secret_name(name: str):
    return f"{name}-regcred"


def create_job_object(image, code: str, name: str, registry_pat: str | None = None):
    # print(client.V1ConfigMapKeySelector(name="main.py"))

    env = prepare_environment(name)

    # Configureate Pod template container
    container = client.V1Container(
        name=name,
        image=image,
        env=env,
        command=["python3"],
        args=["./app/main.py"],
        resources=client.V1ResourceRequirements(
            limits={
                "cpu": "2",
                "ephemeral-storage": "50Gi",
                # "memory": "64Gi",
                "memory": MEMORY,
                "nvidia.com/gpu": "1",
                # "qubernetes.dev/qpu": "1",
            },
            requests={
                "cpu": "2",
                "ephemeral-storage": "0",
                # "memory": "32Gi",
                "memory": MEMORY,
                "nvidia.com/gpu": "1",
                # "qubernetes.dev/qpu": "1",
            },
        ),
        volume_mounts=[
            client.V1VolumeMount(name="app-volume", mount_path="/app", read_only=True)
        ],
    )

    # Create and configurate a spec section
    template = client.V1PodTemplateSpec(
        metadata=client.V1ObjectMeta(labels={"app": name}),
        spec=client.V1PodSpec(
            containers=[container],
            image_pull_secrets=(
                [
                    client.V1LocalObjectReference(
                        name=registry_credentials_secret_name(name)
                    )
                ]
                if registry_pat
                else []
            ),
            runtime_class_name="nvidia",
            restart_policy="Never",
            volumes=[
                client.V1Volume(
                    name="app-volume",
                    config_map=client.V1ConfigMapVolumeSource(name=name),
                )
            ],
        ),
    )

    # Create the specification of deployment
    spec = client.V1JobSpec(template=template)  # , ttl_seconds_after_finished=10

    # Find user name
    # user = whoami()

    # Instantiate the job object
    job_spec = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(
            name=name,
            namespace=NAMESPACE,
            labels={"qubernetes.dev/job.type": "jupyter"},
        ),
        spec=spec,
    )

    job = create_job(job_spec)

    create_config_map_object(code, job, name=name)
    create_environment_secret(name=name)
    if registry_pat:
        create_registry_credentials_secret(
            name=name, image=image, registry_pat=registry_pat
        )

    return job


def create_registry_credentials_secret(
    name: str, image: str, registry_pat: str | None = None
):
    segments = image.split("/")
    # Find user name for images on Docker Hub
    username = segments[0] if len(segments) == 2 else segments[1]
    registry = segments[0] if len(segments) == 3 else "https://index.docker.io/v1/"

    config = {
        "auths": {
            registry: {
                "auth": base64.b64encode(
                    f"{username}:{registry_pat}".encode()
                ).decode(),
            }
        }
    }

    secret = client.V1Secret(
        api_version="v1",
        kind="Secret",
        type="kubernetes.io/dockerconfigjson",
        immutable=True,
        metadata=client.V1ObjectMeta(
            name=registry_credentials_secret_name(name),
            namespace=NAMESPACE,
        ),
        data={
            ".dockerconfigjson": base64.b64encode(
                JSONEncoder().encode(config).encode()
            ).decode(),
        },
    )

    client.CoreV1Api().create_namespaced_secret(namespace=NAMESPACE, body=secret)

    logging.info("Registry credentials created.")


def create_job(job_spec: client.V1Job):
    api_instance = client.BatchV1Api()
    job = api_instance.create_namespaced_job(body=job_spec, namespace=NAMESPACE)
    logging.info("Job created")

    return job


def complete_and_get_job_status(name="qiskit-aer-gpu"):
    job_name = None
    job_completed = False
    api_instance = client.BatchV1Api()

    while not job_completed:
        api_response = api_instance.read_namespaced_job_status(name, NAMESPACE)

        logging.debug("Job status='%s'" % str(api_response.status.start_time))

        if (
            api_response.status.succeeded is not None
            or api_response.status.failed is not None
        ):
            job_completed = True
        elif api_response.status.active is not None:
            if job_name is None:
                job_name = get_pods_in_job(name)

            s = client.CoreV1Api().read_namespaced_pod_status(
                name=job_name,
                namespace=NAMESPACE,
            )

            try:
                if s.status.container_statuses[0].state.terminated is not None:
                    if s.status.container_statuses[0].state.terminated.exit_code == 0:
                        job_completed = True
            except TypeError:
                pass
            finally:
                logging.info("Pod status='%s'" % str(s.status.phase))

        sleep(1)

    return map_job_status_to_stream(api_response.status)


def get_pods_in_job(name="qiskit-aer-gpu"):
    api_instance = client.CoreV1Api()
    api_response = api_instance.list_namespaced_pod(
        NAMESPACE, label_selector=f"app={name}"
    )
    logging.info("Pods in the job='%s'" % str(api_response.items[0].metadata.name))

    return api_response.items[0].metadata.name


def get_job_logs(name="qiskit-aer-gpu"):
    api_response = client.CoreV1Api().read_namespaced_pod_log(
        name=name, namespace=NAMESPACE
    )
    logging.debug("Job logs='%s'" % str(api_response))
    return api_response


def delete_job(name="qiskit-aer-gpu", registry_pat: str | None = None):
    if registry_pat:
        client.CoreV1Api().delete_namespaced_secret(
            registry_credentials_secret_name(name), NAMESPACE
        )
        logging.info("Registry credentials cleanup")

    client.CoreV1Api().delete_namespaced_secret(name, NAMESPACE)
    logging.info("Environment cleanup")

    client.CoreV1Api().delete_namespaced_config_map(
        name,
        NAMESPACE,
        body=client.V1DeleteOptions(propagation_policy="Foreground"),
    )

    api_response = client.BatchV1Api().delete_namespaced_job(
        name,
        NAMESPACE,
        body=client.V1DeleteOptions(propagation_policy="Foreground"),
        _preload_content=False,
    )

    try:
        data = loads(api_response.data)

        logging.info(
            "Job cleanup. status='%s'" % str(data["status"]["conditions"][0]["type"])
        )
    except:
        logging.info(api_response.status)
        logging.info("Job cleanup. status='%s'" % str(api_response.status))

    return api_response


def map_job_status_to_stream(status):
    if status.succeeded is not None:
        return "stdout"
    elif status.failed is not None:
        return "stderr"
    else:
        return "unknown"


def execute(
    code: str, temp_dir: str, docker_image: str, registry_pat: str | None = None
) -> tuple[str, str]:
    # tag = TAG_TEMPLATE.substitute(image=docker_image, version=uuid.uuid4())

    # prepare_build_folder(temp_dir, code)
    # image = docker.buildx.build(
    #     context_path=temp_dir,
    #     tags=[tag],
    #     labels={"qubernetes.cloud/mode": "development"},
    # )

    # logging.info("built new image: %s" % image)

    # docker.image.push(tag)

    id = "".join(random.choices(string.ascii_lowercase + string.digits, k=6))

    name = f"qubernetes-job-{id}"

    create_job_object(
        image=docker_image, code=code, name=name, registry_pat=registry_pat
    )

    stream = complete_and_get_job_status(name=name)

    log = get_job_logs(get_pods_in_job(name=name))

    delete_job(name=name, registry_pat=registry_pat)

    logging.info(stream)

    # delete_directory(temp_dir)

    return log, stream



import os
from q8s.enums import Target
from q8s.project import CacheNotBuiltException, Project, ProjectNotFoundException


def extract_non_none_value(arr):
    non_none_values = [x for x in arr if x is not None]
    return non_none_values[0] if non_none_values else None


def get_docker_image(target: Target = None, logging=None):
    try:
        project = Project()

        image = project.cached_images(target=target)
    except ProjectNotFoundException as e:
        if logging:
            logging.warning(e)
        image = os.environ.get("DOCKER_IMAGE", "vstirbu/benchmark-deps")
    except CacheNotBuiltException as e:
        if logging:
            logging.warning(e)
        image = os.environ.get("DOCKER_IMAGE", "vstirbu/benchmark-deps")
    except Exception as e:
        if logging:
            logging.error(f"Error loading project: {e}")
            logging.warning("Q8Sproject file not found in current folder")
        image = os.environ.get("DOCKER_IMAGE", "vstirbu/benchmark-deps")

    return image


def get_kubeconfig(kubeconfig=None):
    if kubeconfig:
        return kubeconfig

    try:
        project = Project()
        return project.kubeconfig
    except ProjectNotFoundException as e:
        return None



import ast
import sys

from stdlib_list import stdlib_list


class CodeAnalyzer(ast.NodeVisitor):
    stdlibs = []
    version = "{}.{}".format(sys.version_info.major, sys.version_info.minor)

    def __init__(self, code):
        self.imports = set()
        self.stdlibs = self.stdlib_list()

        # with open(file, "r") as source:
        #     tree = ast.parse(source.read())

        #     self.visit(tree)
        tree = ast.parse(code)
        self.visit(tree)

        # print("working with version: ", self.version)

    def stdlib_list(self):
        if sys.version_info.major == 3 and sys.version_info.minor < 10:
            return stdlib_list(self.version)
        else:
            return sys.stdlib_module_names

    def addImport(self, name):
        if name in self.stdlibs:
            pass
        elif name in sys.builtin_module_names:
            pass
        else:
            self.imports.add(name.split(".")[0])

    def visit_Import(self, node):
        for alias in node.names:
            # pprint(vars(alias))
            self.addImport(alias.name)

        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        # pprint(vars(node))
        # pprint(self.file)

        if node.level == 0:
            self.addImport(node.module)
        else:
            if node.module is None:
                for name in node.names:
                    self.addImport(name.asname or name.name)
            else:
                self.addImport(node.module)

        self.generic_visit(node)

    def getImports(self):
        return sorted(self.imports)



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from io import StringIO
import selectors
from subprocess import Popen, PIPE, STDOUT
from os.path import join
from typing import List, Optional
from rich.progress import Progress
import yaml
from dacite import from_dict

from q8s.constants import WORKSPACE
from q8s.enums import Target


def load(path: str):
    """
    Load the project configuration from the Q8Sproject file
    """
    with open(join(path, "Q8Sproject"), "r") as f:
        return yaml.safe_load(f)


def rmdir(directory):
    """
    Recursively remove a directory and its contents
    """
    directory = Path(directory)

    for item in directory.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()

    directory.rmdir()


@dataclass
class Q8SPythonEnv:
    dependencies: List[str]


@dataclass
class Q8STarget:
    python_env: Q8SPythonEnv


@dataclass
class Q8STargets:
    cpu: Optional[Q8STarget]
    gpu: Optional[Q8STarget]

    def keys(self):
        return self.__dataclass_fields__.keys()


@dataclass
class Q8SDocker:
    username: str


@dataclass
class Q8SProject:
    name: str
    python_env: Q8SPythonEnv
    targets: Q8STargets
    docker: Q8SDocker
    kubeconfig: str


class CacheNotBuiltException(Exception):
    pass


class ProjectNotFoundException(Exception):
    pass


class Project:
    name: str
    __path: str
    configuration: Q8SProject
    __images: dict

    def __init__(self, path: str = Path.cwd()):
        try:
            configuration = from_dict(data_class=Q8SProject, data=load(path=path))
        except FileNotFoundError:
            raise ProjectNotFoundException(
                "Q8Sproject file not found in current folder"
            )

        self.configuration = configuration
        self.name = self.configuration.name
        self.__path = path

        self.load_images_cache()

    @property
    def kubeconfig(self):
        return Path(self.configuration.kubeconfig)

    def init_cache(self):
        """
        Initialize the cache directory
        """
        cachepath = join(self.__path, ".q8s_cache")
        Path(cachepath).mkdir(exist_ok=True)

        for target in self.configuration.targets.keys():
            Path(join(self.__path, ".q8s_cache", target)).mkdir(exist_ok=True)

            with open(join(cachepath, target, "requirements.txt"), "w") as f:
                self.__create_requirements_txt(target, f)

            with open(join(cachepath, target, "Dockerfile"), "w") as f:
                self.__create_dockerfile(target, f)

    def check_cache(self):
        result = True

        for target in self.configuration.targets.keys():
            result = self.__check_cache_file(target, "requirements.txt") and result

        return result

    def load_images_cache(self):
        cachepath = join(self.__path, ".q8s_cache", "images")

        if Path(cachepath).exists() is False:
            self.__images = {}
        else:
            with open(cachepath, "r") as f:
                self.__images = yaml.safe_load(f)

    def cached_images(self, target: str) -> str:
        """
        Get the cached images
        """
        cachepath = join(self.__path, ".q8s_cache", "images")

        if Path(cachepath).exists() is False:
            raise CacheNotBuiltException(
                "Images cache not found, build the images first"
            )

        with open(cachepath, "r") as f:
            return yaml.safe_load(f)[target]

    def build_container(
        self, target: str, progress: Progress, silent: bool, push: bool = True
    ):
        """
        Build the container image
        """
        targetpath = join(self.__path, ".q8s_cache", target)

        task = progress.add_task(
            description=f"[cyan]Building container for {target}...", total=1
        )

        # start the docker build command in subprocess and capture the output
        build_process = Popen(
            [
                "docker",
                "build",
                "--progress",
                "plain",
                "--tag",
                self.__image_name(target),
                targetpath,
            ],
            stdout=PIPE,
            stderr=STDOUT,
            bufsize=1,
            universal_newlines=True,
        )

        def handle_output(stream, mask):
            # Because the process' output is line buffered, there's only ever one
            # line to read when this function is called
            line = stream.readline()
            if not silent:
                progress.console.print(line, end="")

        # Register callback for an "available for read" event from subprocess' stdout stream
        selector = selectors.DefaultSelector()
        selector.register(build_process.stdout, selectors.EVENT_READ, handle_output)

        # Loop until subprocess is terminated
        while build_process.poll() is None:
            # Wait for events and handle them with their registered callbacks
            events = selector.select()
            for key, mask in events:
                callback = key.data
                callback(key.fileobj, mask)

        selector.close()

        if build_process.returncode != 0:
            progress.advance(task)
            raise Exception("Failed to build the container")
        else:
            progress.console.print(f"Container {self.__image_name(target)} built")
            progress.advance(task, 1)

        if push:
            self.push_container(target, progress, silent)

        self.__images[target] = self.__image_name(target)

    def push_container(self, target: str, progress: Progress, silent: bool):
        """
        Push the container image to the registry
        """
        task = progress.add_task(
            description=f"[cyan]Pushing container for {target}...", total=1
        )

        push_process = Popen(
            ["docker", "push", self.__image_name(target)],
            stdout=PIPE,
            stderr=STDOUT,
            bufsize=1,
            universal_newlines=True,
        )

        def handle_output(stream, mask):
            # Because the process' output is line buffered, there's only ever one
            # line to read when this function is called
            line = stream.readline()
            if not silent:
                progress.console.print(line, end="")

        # Register callback for an "available for read" event from subprocess' stdout stream
        selector = selectors.DefaultSelector()
        selector.register(push_process.stdout, selectors.EVENT_READ, handle_output)

        # Loop until subprocess is terminated
        while push_process.poll() is None:
            # Wait for events and handle them with their registered callbacks
            events = selector.select()
            for key, mask in events:
                callback = key.data
                callback(key.fileobj, mask)

        selector.close()

        if push_process.returncode != 0:
            progress.advance(task)
            raise Exception("Failed to push the container")
        else:
            progress.advance(task)

    def update_images_cache(self):
        """
        Update the images cache
        """
        with open(join(self.__path, ".q8s_cache", "images"), "w") as f:
            yaml.dump(self.__images, f)

    def clear_cache(self):
        """
        Clear the cache directory
        """
        cachepath = join(self.__path, ".q8s_cache")
        rmdir(cachepath)

    def __docker_login(self) -> str:
        return self.configuration.docker.username

    def __image_name(self, target: str):
        return f"{self.__docker_login()}/q8s-{self.name.lower()}:{target}"

    def __check_cache_file(self, target: str, file: str):
        cachepath = join(self.__path, ".q8s_cache", target, file)
        if Path(cachepath).exists() is False:
            print(f"Cache file {cachepath} does not exist")
            return False

        file = StringIO()
        self.__create_requirements_txt(target, file)

        with open(cachepath, "r") as f:
            if file.getvalue() != f.read():
                print(f"Cache file {cachepath} is outdated")
                return False

        return True

    def __create_requirements_txt(self, target: str, f):
        print("# This file is autogenerated by q8sctl", file=f)
        print("# Do not edit manually", file=f)

        print("\n# Common dependencies:", file=f)

        for dep in self.configuration.python_env.dependencies:
            print(f"{dep}", file=f)

        print("\n# Target specific dependencies:", file=f)

        for dep in self.__get_target(target=target).python_env.dependencies:
            print(f"{dep}", file=f)

    def __create_dockerfile(self, target: str, f):
        print("# This file is autogenerated by q8sctl", file=f)
        print("# Do not edit manually", file=f)

        base_images = {
            "cpu": "--platform=linux/amd64 python:3.12-slim",
            "gpu": "vstirbu/q8s-cuda12:latest",
        }
        print(f"\nFROM {base_images[target]}", file=f)
        print("", file=f)

        print(
            f"LABEL org.opencontainers.image.created={datetime.now().isoformat()}",
            file=f,
        )
        print(f"LABEL org.opencontainers.image.title={self.name}", file=f)
        print("", file=f)

        print(f"WORKDIR {WORKSPACE}", file=f)
        print("COPY requirements.txt .", file=f)
        print("RUN pip install -r requirements.txt", file=f)

        if target == "gpu":
            print("", file=f)
            print("RUN ln -s /usr/bin/python3 /usr/bin/python", file=f)

    def __get_target(self, target: str) -> Q8STarget:
        if hasattr(self.configuration.targets, target) is False:
            raise Exception(f"Target {target} not found")

        return getattr(self.configuration.targets, target)



import unittest
from unittest.mock import patch, MagicMock
from q8s.matplotlib.backend import Q8SLoggerBackend

import matplotlib.pyplot as plt


class TestQ8SLoggerBackend(unittest.TestCase):
    @patch("builtins.print")
    def test_print_png_logs_base64(self, mock_print):
        # Create a simple plot
        fig, ax = plt.subplots()
        ax.plot([0, 1], [0, 1])

        # Create an instance of the custom backend
        canvas = Q8SLoggerBackend(fig)

        # Call the print_png method
        canvas.print_png(None)

        # Check that print was called with the expected base64 string
        self.assertTrue(mock_print.called)
        args, _ = mock_print.call_args
        self.assertTrue(args[0].startswith("data:image/png;base64,"))


if __name__ == "__main__":
    unittest.main()
